<?php

/**
 * Create object to parse a pdf file.
 * @author Marius Walk
 */
class RHPDFParser {

	//array with all citation lines
	private $citations;

	/**
	 * Creates a RHPDFParser object.
	 * @return RHPublication
	 * @param $url of the pdf file
	 */
	function __construct($url) {
		// Parse pdf file and build necessary objects.
		$parser = new \Smalot\PdfParser\Parser();
		$pdf    = $parser->parseFile($url);
		//extract text from pdf
		$text = $pdf->getText();
		$references = $this->getReferenceSection($text);
		$regex = $this->getCitationStyle($references);
		$results = preg_split($regex, $references);
		foreach ($results as $key => $citation) {
			$results[$key] = substr($citation, 0, 600);
		}
		$this->citations = $results;
		
		return $this;
	}

	/**
	 * Gets the citations.
	 * @return array with citations
	 */
	public function getCitations() {
		return $this->citations;
	}

	/**
	 * Gives back the text section where the references are stated.
	 * @return $references string with reference section or fals if not found
	 */
	private function getReferenceSection ($text) {
		$section_terms = array('References', 'Bibliography', 'Literatur', 'Literaturverzeichnis');
		$position_references = 0;
		for ($i=0; $position_references == 0 && $i < count($section_terms); $i++) {
			$position_references = strripos($text, $section_terms[$i]);
		}
		if ($position_references != 0) {
			$references = substr($text, $position_references);
			$references = substr($references, strlen($section_terms[$i]) + 1);
			return $references;
		} else {
			return $text;
		}
	}

	/**
	 * Gives back a regex of the most used citation style.
	 * @return $regex
	 */
	private function getCitationStyle($references) {
		$styles = array('@\[[0-9]{1,3}\]@' => 0,
			'@\.\n@' => 0,
			'@\. \n@' => 0);
		foreach ($styles as $regex => $quantity) {
			$styles[$regex] = preg_match_all($regex, $references);
		}
		arsort($styles);
		$keys = array_keys($styles);
		return array_shift($keys);
	}

	/**
	 * Parses each citation with the help of FreeCite (http://freecite.library.brown.edu/).
	 * Gives back a list of RHPublications (one for each citation).
	 * @return array of RHPublication objects
	 */
	public function getCitatedPublications() {
		$publications = array();
		//URL to FreeCite API
		$url = 'http://freecite.library.brown.edu/citations/create';
		//data contains the parameters sent to FreeCite API
		$data = '';
		$citations = $this->citations;
		foreach ($citations as $citation) {
			$data .= 'citation[]='.str_replace('%00', '', urlencode($citation)).'&';
		}
		$data = substr($data, 0, -1);
		//echo '<p>Ergebnis XML Freecite:</p>'.$data.'<hr/>';
		$header = array('Accept: text/xml');
		//Execute HTTP-Request at FreeCite
		$ch = curl_init($url);
		curl_setopt($ch, CURLOPT_HTTPHEADER, $header);
		curl_setopt($ch, CURLOPT_POSTFIELDS, $data);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		$response = curl_exec($ch);
		//echo '<p>Ergebnis XML Freecite:</p>'.$response.'<hr/>';
		curl_close($ch);

		//Build xml-Object from HTTP-response
		$xml = @simplexml_load_string($response);
		if ($xml !== false && isset($xml->citation)) {
			//echo '<p>Ergebnisse von FreeCite:</p>';
			$citations = $xml->citation;
			foreach ($citations as $citation) {
				$pub = $this->buildRHPublicationFromCitationXML($citation);
				//echo $pub->toString();
				//echo '<hr/>';
				array_push($publications, $pub);
			}
		} else {
			//echo '<p>Ergebnisse von FreeCite nicht lesbar</p><hr/>';
			foreach ($this->citations as $citation) {
				$pub = new RHPublication('', array(), $citation);
				array_push($publications, $pub);
			}
		}
		return $publications;
	}

	/**
	 * Gives back a RHPublication object from xml structured citaion element.
	 * @return RHPublication
	 */
	private function buildRHPublicationFromCitationXML($citation) {
			$authors = array();
			$authors_xml = $citation->authors->author;
			foreach ($authors_xml as $author) {
				array_push($authors, (string) $author);
			}
			$title = (string) $citation->title;
			$raw_string = (string) $citation->raw_string;
			$pub = new RHPublication($title, $authors, $raw_string);

			$extra_values = array();
			if (isset($citation->pages)) {
				$extra_values['pages'] = (string) $citation->pages;
			}
			if (isset($citation->year)) {
				$extra_values['year'] = (string) $citation->year;
			}
			if (isset($citation->volume)) {
				$extra_values['volume'] = (string) $citation->volume;
			}
			$pub->setValues($extra_values);
		return $pub;
	}
	
}