<?php


/**
 * Represents a publication.
 * @author Marius Walk
 */
class RHPublication {
	
	//all attribute declarations
	public $attributes = array('title', 'authors', 'university', 'department',
								'conference', 'keyword', 'abstract', 'rating',
								'url', 'publicationType', 'number', 'volume', 'year',
								'pages', 'citation', 'doi', 'journal');

	//title of the publication
	private $title;

	private $citation;

	private $authors; //has to be an array!!!

	private $university;

	private $department;

	private $conference;

	private $keyword;

	private $abstractOfPublication;

	private $rating;

	private $url;

	private $publicationType;

	private $number;

	private $volume;

	private $year;

	private $pages;

	private $doi;

	private $journal;


	/**
	 * Creates a publication object.
	 * @return RHPublication
	 * @param $title the title of the publication
	 * @param $authors an array with all authors
	 */
	function __construct($title, $authors, $citation) {
		$this->title = $title;
		$this->authors = $authors;
		$this->citation = $citation;
		return $this;
	}

	/**
	 * Get the title.
	 * @return title string
	 */
	public function getTitle() {
		return $this->title;
	}

	/**
	 * Get the authors.
	 * @return authors[]
	 */
	public function getAuthors() {
		return $this->authors;
	}


	/**
	 * Get the citation line.
	 * @return citation (string)
	 */
	public function getCitation() {
		return $this->citation;
	}

	/**
	 * Creates an array with all set values.
	 * @return array (for example: array['title'] = 'title of the publication')
	 */
	public function getValues() {
		$values = array();
		foreach ($this->attributes as $attribute) {
			if (isset($this->$attribute) && $this->$attribute != '') {
				$values[$attribute] = $this->$attribute;
			}
		}
		return $values;
	}

	/**
	 * Sets the given Values.
	 * @param $values: array (for example: array['title'] = 'title of the publication')
	 */
	public function setValues($values) {
		foreach ($values as $attribute => $value) {
			//test if it $attribute is a valid RHPublication attribute
			if (in_array($attribute, $this->attributes)) {
				$this->$attribute = $value;
			}
		}
	}

	/**
	 * Adds the name of an author.
	 * @param $name: string
	 */
	public function addAuthor($name) {
		if (is_array($this->authors)) {
			array_push($this->authors, $name);
		} else {
			$this->authors = array($name);
		}
	}

	/**
	 * To be valid author or title of a RHPublication can't be empty.
	 * @return boolean
	 */
	public function isValid() {
		if (($this->title != "") && (isset($this->title)) && (isset($this->authors))) {
			if (is_array($this->authors)) {
				foreach ($this->authors as $author) {
					if ($author != "") {
						return true;
					}
				}
			}
		}
	return false;
	}

	/**
	 * Compares to RHPublications.
	 * @param $publication RHPublication object
	 * @return true if the objects represent the same publication
	 */
	private function compare($publication) {
		//do the publications have the same title?
		$clear_citation_a = RHHelper::cleanUp($this->citation); //citation PubA
		$clear_title_b = RHHelper::cleanUp($publication->getTitle()); //citation PubA
		if (strpos($clear_citation_a, $clear_title_b) === false) {
			return false;
		}
		//do the publications have the same author?
		$lengths = array();
		foreach ($publication->getAuthors() as $author) {
			//find last name
			$names = explode(' ', $author);
			foreach ($names as $name) {
				$lengths[$name] = strlen($name);
			}
			arsort($lengths);
			$names = array_keys($lengths);
			$last_name = array_shift($names);
			$clear_last_name = RHHelper::cleanUp($last_name);
			if (@strpos($clear_citation_a, $clear_last_name) !== false) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Return Html String to present the object
	 * @return string
	 */
	public function toString() {
		$return = '';
		foreach ($this->getValues() as $key => $value) {
			if ($key == 'authors') {
				$return .= '<p>Autoren: ';
				foreach ($value as $author) {
					$return .= ''.$author.', ';
				}
				$return .= '</p>';
			} else {
				$return .= '<p>'.$key.': '.$value.'</p>';
			}
		}
		return $return;
	}

	/**
	 * Searches the given services for more information about this publication.
	 * @param $services array wiht several service object (for exampl RHMendeley object)
	 */
	public function getInfoFromServices($services) {
		$results = array(); //array which holds search results as RHPublication objects
		foreach ($services as $service) {
			$results = $service->getSearchResults($this);
			//check if a search results matches this publication
			foreach ($results as $pub) {
				//if it matches fill in the values from the search result
				if ($this->compare($pub)) {
				 	$search_values = $pub->getValues();
				 	$this->setValues($search_values);
				 	break;
				}
			} 
		}
	}

	/**
	 * Creates a page for this publication in the wiki.
	 * @return boolean success
	 */
	public function saveAsWikiPage() {
		global $wgUser;

		//create WikiPage Object
		$title = Title::newFromText($this->getTitle());
		$wikiPage = WikiPage::factory( $title );

		$db = wfGetDB( DB_SLAVE );
		wfProfileIn( __METHOD__ );

		$title = Title::newFromText($this->title, $defaultNamespace=NS_MAIN);

		//get variables for used template
		$variable_settings = $db->select(
			'reference_helper',
			array('user', 'standard'),
			'field_type = "variable"',
			__METHOD__,
			array( 'ORDER BY' => 'field_type' )
		);
		//get current template settings
		$template_settings = $db->select(
			'reference_helper',
			'user',
			'field_type = "template"',
			__METHOD__,
			array( 'ORDER BY' => 'field_type' )
		);
		$template = '';
		if ($template_settings->numRows() == 1) {
			$template = $template_settings->fetchRow()['user'];
		}
		//create wikitext
		$page_text = '{{'.$template.'';
		foreach ($variable_settings as $variable) {
			$standard = $variable->standard;
			if (isset($this->$standard)) {
				if ($standard == 'authors') {
					$authors = '|'.$variable->user.'=';
					foreach ($this->$standard as $author) {
						$authors .= ''.$author.',';
					}
					$authors = substr($authors, 0, -1);
					$authors = str_replace(', ', ',', $authors);
				} else {
					$page_text .= '|'.$variable->user.'='.$this->$standard.'';
				}
			}
		}
		$page_text .= $authors;
		$page_text .= '}}';
		$content = new WikitextContent($page_text);
		//save wiki page
		$wikiPage->doEditContent( $content, 'created by ReferenceHelper', 'EDIT_NEW', false,
								   $wgUser, null );

		return true;
	}
	
}